"""Unit-test the REST-API client."""

from unittest import mock

import pytest
import requests
import requests.exceptions

from .. import rest


class ClientTest:
    """Unit-tests for the REST-API client."""

    @staticmethod
    @mock.patch("moflask.requests.sessions.Session.request")
    def test_base_url_restriction(mock_request):
        """Test that requests to a different base URL are denied.

        This is useful for APIs that provide URLs to other objects in their responses. This way
        we can ensure that no data is sent to an unintended location, even if we follow such
        untrusted URLs.
        """
        client = rest.Client("https://base.example.org")
        expected_message = "This client only sends requests to https://base.example.org"
        with pytest.raises(requests.exceptions.URLRequired, match=expected_message):
            client.get(url="https://other.example.org")
        assert not mock_request.called

    @staticmethod
    @mock.patch("requests.Session.send")
    def test_auth_middleware(mock_send):
        """Test that the a configured auth middleware is called and can manipulate the request."""

        def middleware(request: requests.PreparedRequest):
            request.headers["X-Middleware-Test"] = "success"
            return request

        client = rest.Client("https://base.example.org", auth=middleware)
        client.get("path")

        assert mock_send.called
        _, args, _ = mock_send.mock_calls[0]
        request = args[0]
        assert request.headers["X-Middleware-Test"] == "success"



from ipaddress import ip_address

_split = lambda x: x.split(",") if x else []


class ProxyFix(object):
    """This is a slightly modified version of werkzeug's ProxyFix.

    Instead of using a fixed number of proxies it uses a list of trusted IP
    addresses.

    This middleware can be applied to add HTTP proxy support to an
    application that was not designed with HTTP proxies in mind.  It
    sets `REMOTE_ADDR`, `HTTP_HOST` from `X-Forwarded` headers.

    The original values of `REMOTE_ADDR` and `HTTP_HOST` are stored in
    the WSGI environment as `werkzeug.proxy_fix.orig_remote_addr` and
    `werkzeug.proxy_fix.orig_http_host`.

    :param app: the WSGI application
    :param num_proxies: the number of proxy servers in front of the app.
    """

    def __init__(self, app):
        self.wsgi = app.wsgi_app
        proxies = app.config.get("PROXYFIX_TRUSTED", ["127.0.0.1"])
        self.trusted = frozenset(ip_address(p.strip()) for p in proxies)

    def get_remote_addr(self, forwarded_for, remote):
        """Selects the new remote addr from the given list of ips in
        X-Forwarded-For. It picks the "rightmost" untrusted IP address.
        """
        for ip_str in reversed([remote] + forwarded_for):
            ip_str = ip_str.strip()
            try:
                if not ip_address(ip_str) in self.trusted:
                    return ip_str
            except ValueError:
                return ip_str
        return remote

    def update_environ(self, environ):
        env = environ.get
        remote_addr = env("REMOTE_ADDR")
        if not remote_addr:
            return

        try:
            remote_addr_ip = ip_address(remote_addr)
        except ValueError:
            remote_addr_ip = ip_address("127.0.0.1")

        forwarded_proto = env("HTTP_X_FORWARDED_PROTO", "")
        forwarded_for = _split(env("HTTP_X_FORWARDED_FOR", ""))
        forwarded_host = env("HTTP_X_FORWARDED_HOST", "")
        environ.update(
            {
                "werkzeug.proxy_fix.orig_wsgi_url_scheme": env("wsgi.url_scheme"),
                "werkzeug.proxy_fix.orig_remote_addr": env("REMOTE_ADDR"),
                "werkzeug.proxy_fix.orig_http_host": env("HTTP_HOST"),
            }
        )

        if remote_addr_ip in self.trusted:
            if forwarded_host:
                environ["HTTP_HOST"] = forwarded_host
            if forwarded_proto:
                https = "https" in forwarded_proto.lower()
                environ["wsgi.url_scheme"] = "https" if https else "http"

        remote_addr = self.get_remote_addr(forwarded_for, remote_addr)
        if remote_addr is not None:
            environ["REMOTE_ADDR"] = remote_addr

    def __call__(self, environ, start_response):
        self.update_environ(environ)
        return self.wsgi(environ, start_response)



from copy import copy
from unittest import TestCase

from ..wsgi import ProxyFix


class ProxyFixTest(TestCase):
    def app_stub(self, trusted):
        class Object(object):
            wsgi_app = None

        app = Object()
        app.config = {}
        app.config["PROXYFIX_TRUSTED"] = trusted
        return app

    def _copy_env(self, env):
        e = copy(env)
        e["werkzeug.proxy_fix.orig_http_host"] = env["HTTP_HOST"]
        e["werkzeug.proxy_fix.orig_remote_addr"] = env["REMOTE_ADDR"]
        e["werkzeug.proxy_fix.orig_wsgi_url_scheme"] = env["wsgi.url_scheme"]
        return e

    def test_one_forwarded_for_layer(self):
        fix = ProxyFix(self.app_stub(["127.0.0.1"]))
        env = {
            "REMOTE_ADDR": "127.0.0.1",
            "HTTP_HOST": "example.com",
            "wsgi.url_scheme": "http",
            "HTTP_X_FORWARDED_FOR": "8.8.8.8",
            "HTTP_X_FORWARDED_PROTO": "https",
        }
        e = self._copy_env(env)
        e["REMOTE_ADDR"] = "8.8.8.8"
        e["wsgi.url_scheme"] = "https"
        fix.update_environ(env)
        self.assertEqual(e, env)

    def test_multiple_trusted_multiple_untrusted(self):
        fix = ProxyFix(self.app_stub(["127.0.0.1", "10.0.0.1"]))
        rm = fix.get_remote_addr(
            ["8.8.8.8", "10.0.0.1", "127.0.0.1"],
            "1.1.1.1",
        )
        self.assertEqual("8.8.8.8", rm)

    def test_all_trusted(self):
        fix = ProxyFix(self.app_stub(["127.0.0.1"]))
        self.assertEqual("127.0.0.1", fix.get_remote_addr([], "127.0.0.1"))

    def test_no_trusted_layer(self):
        fix = ProxyFix(self.app_stub(["127.0.0.1"]))
        env = {
            "REMOTE_ADDR": "8.8.8.8",
            "HTTP_HOST": "example.com",
            "wsgi.url_scheme": "http",
            "HTTP_X_FORWARDED_FOR": "8.8.8.8",
            "HTTP_X_FORWARDED_HOST": "untrusted.com",
            "HTTP_X_FORWARDED_PROTO": "https",
        }
        e = self._copy_env(env)
        fix.update_environ(env)
        self.assertEqual(e, env)

    def test_new_ip_equals_old_ip(self):
        fix = ProxyFix(self.app_stub(["127.0.0.1"]))
        env = {
            "REMOTE_ADDR": "127.0.0.1",
            "HTTP_HOST": "example.com",
            "wsgi.url_scheme": "http",
            "HTTP_X_FORWARDED_FOR": "127.0.0.1",
            "HTTP_X_FORWARDED_PROTO": "https",
        }
        e = self._copy_env(env)
        e["wsgi.url_scheme"] = "https"
        fix.update_environ(env)
        self.assertEqual(e, env)

    def test_no_proxy_works_transparently(self):
        fix = ProxyFix(self.app_stub(["127.0.0.1"]))
        env = {"REMOTE_ADDR": "127.0.0.1", "HTTP_HOST": "example.com", "wsgi.url_scheme": "http"}
        e = self._copy_env(env)
        fix.update_environ(env)
        self.assertEqual(e, env)



"""Auth middlewares for the requests library."""

import urllib.parse

import flask
import requests

from . import rest


class AuthAppClient(rest.Client):
    """A REST client for retrieving JWTs from the auth-app."""

    @classmethod
    def from_app(cls, app=None):
        """Create a new instance by reading the config from the Flask app config."""
        app = app or flask.current_app
        return cls(
            app.config["IMPACT_STACK_AUTH_APP_URL"],
            app.config["IMPACT_STACK_AUTH_APP_KEY"],
        )

    def __init__(self, base_url, api_key):
        """Create a new client instance."""
        super().__init__(base_url)
        self.api_key = api_key

    def get_token(self, organization):
        """Get a JWT token for a sub-organization."""
        organization = urllib.parse.quote_plus(organization)
        response = self.post("token", organization, json=self.api_key)
        return response.json()["token"]


class AuthAppMiddleware:
    """Middleware for authenticating using JWT tokens.

    The middleware transparently requests an API-token from the auth-app on-demand.
    """

    def __init__(self, organization, client=None):
        """Create new auth-app requests auth middleware."""
        self.client = client or AuthAppClient.from_app()
        self.organization = organization

    def get_token(self):
        """Get a JWT token from the auth-app."""
        return self.client.get_token(self.organization)

    def __call__(self, request: requests.PreparedRequest):
        """Add the JWT token to the request."""
        request.headers["Authorization"] = "Bearer " + self.get_token()
        return request



"""Test flask_security related loggging.

This needs some mocking of sys.modules, so isolate this in it's own test
file.
"""
# pylint: disable=no-member,protected-access

import logging as _logging

from flask import Flask
from flask_login import LoginManager

from moflask import logging


def test_user_context_filter():
    """Test UserContext enriches the LogRecord.

    Pulls information from a Flask-Security `currrent_user` object.
    """
    app = Flask(__name__)

    # Provide a minimal setup for `current_user` to work
    login_manager = LoginManager()
    login_manager.init_app(app)
    login_manager.user_loader(lambda user_id: None)

    context_filter = logging.add_user_data
    record = _logging.makeLogRecord({})
    with app.test_request_context("/"):
        context_filter(record)
    assert record.user
    assert record.user == "<AnonymousUser>"
